<?php
	set_time_limit(120);
	ini_set('memory_limit', '128M');
	chdir(__DIR__);
	
	//define('MAXLEN', 5000000);
	define('MAXLEN', 25000000);
	//define('MAXLEN', 50000000);
	define('RET250', "250 OK\r\n");
	define('RET500', "500 Syntax error, command unrecognized\r\n");
	
	// Extract the host from the from address
	$host = 'example.com';
	@list($from) = file('config.from.txt', FILE_IGNORE_NEW_LINES);
	if ($from && preg_match('/<[^@>]+@([^>]+)>/', $from, $m)) {
		list(, $host) = $m;
	}
	
	// forwards a whole domain (@example.com) to another SMTP server
	$forward = array();
	$fp = fopen('config.domainforward.txt', 'r');
	while ((@list($from, $to) = fgetcsv($fp, null, ' ')) !== false) {
		$forward[$from] = $to;
	}
	fclose($fp);
	
	// forwards e-mail addresses to other addresses
	$alias = array();
	$fp = fopen('config.forward.txt', 'r');
	while ((@list($from, $to) = fgetcsv($fp, null, ' ')) !== false) {
		$alias[$from] = $to;
	}
	fclose($fp);
	
	// local mailboxes
	$local = array_map('basename', glob('mbox/*/', GLOB_ONLYDIR));
	
	$from = '';
	$tofwd = $towhom = array();
	
	echo "220 Philippe SMTP Server\r\n";
	while (($line = fgets(STDIN)) !== false) {
		preg_match('/^(....)( (.*))?$/', substr($line, 0, -2), $out);
		@list(, $command, , $argument) = $out;
		switch (strtoupper($command)) {
			case 'HELO':
				echo "250 Hello\r\n";
				break;
			case 'MAIL':
				if (!preg_match('/^FROM:<(.*)>$/i', $argument, $mail))
					echo RET500;
				else {
					@list(, $from) = $mail;
					$from = strtolower($from);
					echo RET250;
				}
				break;
			case 'RCPT':
				if (!preg_match('/^TO:<(.*)>$/i', $argument, $rcpt))
					echo RET500;
				else {
					@list(, $to) = $rcpt;
					$to = strtolower($to);
					if (in_array($to, $local)) {
						$towhom[] = $to;
						echo RET250;
					} else if (isset($alias[$to])) {
						$to = $alias[$to];
						@list(, $domain) = explode('@', $to, 2);
						if (getmxrr($domain, $hosts))
							$tofwd[$hosts[0]][] = $to;
						echo "251 User not local; will forward to <$to>\r\n";
					} else {
						@list(, $domain) = explode('@', $to, 2);
						if (isset($forward[$domain])) {
							$tofwd[$forward[$domain]][] = $to;
							echo "251 User not local; will forward\r\n";
						} else
							echo "550 No such user here\r\n";
					}
				}
				break;
			case 'DATA':
				echo "354 Start mail input; end with <CRLF>.<CRLF>\r\n";
				$origdata = $data = '';
				while (($line = fgets(STDIN)) !== false) {
					if ($line === ".\r\n")
						break;
					$origdata .= $line;
					if (@$line[0] === '.')
						$line = substr($line, 1);
					$data .= $line;
				}
				if (strlen($data) > MAXLEN) {
					echo "552 Email too big\r\n";
					break;
				} else {
					// https://www.jwz.org/doc/content-length.html
					$head = "From <$from> " . date('Y-m-d H:i:s') . "\r\n";
					$data = $head . str_replace("\r\nFrom ", "\r\n>From ", $data) . "\r\n";
					foreach ($towhom as $to) {
						file_put_contents('mbox/' . $to . '.txt', $data, LOCK_EX | FILE_APPEND);
						file_put_contents('mbox/' . $to . '/' . uniqid('', true) . '.eml', $data);
					}
					foreach ($tofwd as $domain => $tos) {
						if ($fp = stream_socket_client('tcp://' . $domain . ':25', $eno, $estr, 30)) {
							fgets($fp);
							fwrite($fp, "HELO $host\r\n");
							fgets($fp);
							fwrite($fp, "MAIL FROM:<$from>\r\n");
							fgets($fp);
							foreach ($tos as $to) {
								fwrite($fp, "RCPT TO:<$to>\r\n");
								fgets($fp);
							}
							fwrite($fp, "DATA\r\n");
							fgets($fp);
							fwrite($fp, $origdata);
							fwrite($fp, ".\r\n");
							fgets($fp);
							fwrite($fp, "QUIT\r\n");
							fgets($fp);
							fclose($fp);
						}
					}
				}
				$from = '';
				$tofwd = $towhom = array();
				echo RET250;
				break;
			case 'RSET':
				$from = '';
				$tofwd = $towhom = array();
				echo RET250;
				break;
			case 'NOOP':
				echo RET250;
				break;
			case 'QUIT':
				echo "221 Philippe SMTP Server Closing\r\n";
				break 2;
			default:
				echo RET500;
				break;
		}
	}
?>
