<?php
	require 'mdp.php';
	
	function txt2file($txt) {
		return substr($txt, 0, -4);
	}
	$files = array_map('txt2file', array_map('basename', glob('../mbox/*.txt', GLOB_BRACE)));
	
	$file = @$_GET['file'] ?: @$files[0];
	if (!in_array($file, $files)) {
		die('No file.');
	}
	$idx = '../idx/' . $file . '.idx';
	if (!file_exists($idx)) {
		touch($idx);
	}
	$suffixsent = '&file=' . rawurlencode($file);
	$mbox = '../mbox/' . $file . '.txt';
	
	function philsize($len) {
		return round($len / 1024, 2) . ' kB';
	}
	
	// Content-Id
	$urls = array();
	
	function html_img_cid($html) {
		global $urls;
		$doc = new DOMDocument('1.0', 'UTF-8');
		$doc->formatOutput = true;
		$doc->preserveWhiteSpace = false;
		@$doc->loadHTML($html);
		foreach ($doc->getElementsByTagName('img') as $tag) {
			@list($scheme, $value) = explode(':', $tag->getAttribute('src'), 2);
			if ($scheme === 'cid' && isset($urls[$value = '<' . $value . '>']))
				$tag->setAttribute('src', $urls[$value]);
		}
		return $doc->saveHTML();
	}
	
	function myfrom2mime($from) {
		if (substr($from, 0, 5) === 'From ')
			$from = substr($from, strpos($from, "\r\n") + 2, -2);
		return $from;
	}
	
	define('TYPETOKEN', '[A-Za-z0-9.$%*_+~`\'-]+');
	function mytypeparse($string, $noslash = false) {
		//$string = strtolower($string);
		$type = '';
		$args = array();
		// https://www.cs.columbia.edu/sip/syntax/rfc2543.html#media-type
		if (preg_match('@^\s*(' . TYPETOKEN . ($noslash ? '' : '/' . TYPETOKEN) . ')@', $string, $m)) {
			list($start, $type) = $m;
			$string = substr($string, strlen($start));
			while ($string && preg_match('@^\s*;\s*(' . TYPETOKEN . ')=(' . TYPETOKEN . '|"([^"]|\\")+")@', $string, $m)) {
				list($start, $key, $value) = $m;
				if (@$value[0] === '"')
					$value = substr(str_replace('\\"', '"', $value), 1, -1);
				//$args[$key] = $value;
				$args[strtolower($key)] = $value;
				$string = substr($string, strlen($start));
			}
		}
		//return array($type, $args);
		return array(strtolower($type), $args);
	}
	
	$has_html = false;
	function mymimebody($head, $body, $dataurl) {
		global $urls;
		global $has_html;
		
		if (strcasecmp(@$head['Content-Transfer-Encoding'], 'quoted-printable') == 0) {
			$body = preg_replace_callback('/=([A-F0-9][A-F0-9])/', function($m) { return chr(hexdec($m[1])); }, $body);
			$body = str_replace("=\r\n", '', $body);
		}
		if (strcasecmp(@$head['Content-Transfer-Encoding'], 'base64') == 0) {
			$body = base64_decode(str_replace("\r\n", '', $body));
		}
		@list($type, $args) = mytypeparse(@$head['Content-Type'] ?: 'text/plain');
		if (preg_match('@^multipart/(mixed|alternative|related)$@', $type, $m) && isset($args['boundary'])) {
			list(, $multipart) = $m;
			$boundary = $args['boundary'];
			
			$pos = strpos($body, "\r\n--$boundary--\r\n");
			if ($pos !== false)
				$body = substr($body, 0, $pos);
			$parts = explode("\r\n--$boundary\r\n", "\r\n" . $body);
			array_shift($parts);
			
			$html = '';
			if ($multipart === 'related') {
				$first = array_shift($parts);
				if ($first) {
					//$urls = array();
					foreach ($parts as $part) {
						list($h, $b) = myparsemime($part, true, true);
						if (isset($h['Content-Id']))
							$urls[$h['Content-Id']] = $b;
					}
					list(, $html) = myparsemime($first, true, false);
				}
			} else {
				foreach ($parts as $part) {
					list(, $body) = myparsemime($part, true);
					if ($multipart === 'mixed')
						$html .= $body;
					else if ($multipart === 'alternative' && $body)
						$html = $body;
				}
			}
			return $html;
		} else {
			
			@list($dispo, $dispoargs) = mytypeparse($head['Content-Disposition'], true);
			if ($dispo === 'attachment' && isset($dispoargs['filename']) && substr($type, 0, 6) !== 'image/') {
				$url = 'data:' . $type . ';base64,' . base64_encode($body);
				return '<div class=article><p><a download="' . htmlspecialchars($dispoargs['filename']) . '" href="' . htmlspecialchars($url) . '">' . htmlspecialchars($dispoargs['filename']) . "</a></div>\n";
			}
			
			if (preg_match('@^text/(plain|html)$@', $type, $m)) {
				list(, $text) = $m;
				if (!$charset = @$args['charset'])
					$charset = 'UTF-8';
				$body = @iconv($charset, 'UTF-8', str_replace("\r\n", "\n", $body));
				if ($text === 'html') {
					$has_html = true;
					if (isset($_GET['html'])) {
						if ($urls)
							$body = html_img_cid($body);
						$url = 'data:text/html;charset=utf-8;base64,' . base64_encode($body);
						if ($dataurl)
							return $url;
						return "<p>HTML</p>\n<div class=article>\n<p><iframe height=500 sandbox src=\"" . htmlspecialchars($url) . "\" width=100%></iframe></div>\n";
					}
				} else if ($text === 'plain') {
					if ($dataurl)
						return 'data:text/plain;charset=utf-8;base64,' . base64_encode($body);
					return "<pre class=article>\n" . htmlspecialchars($body) . "</pre>\n";
				}
			} else if (preg_match('@^(image/[^;]+)$@', $type, $m)) {
				list(, $ct) = $m;
				$url = 'data:' . $ct . ';base64,' . base64_encode($body);
				if ($dataurl)
					return $url;
				return '<div class=article><p><img src="' . htmlspecialchars($url) . '" /></div>' . "\n";
			}
		}
		return '';
	}
	
	function myparsemime($mime, $withbody = false, $dataurl = false) {
		$head = array();
		$lines = explode("\r\n", $mime);
		foreach ($lines as $i => $line) {
			if ($line === "")
				break;
			if (@$line[0] === ' ' || @$line[0] === "\t") {
				if ($head)
					$head[count($head) - 1][1] .= @iconv_mime_decode($line, 0, 'UTF-8');
			} else {
				//@list($key, $value) = explode(': ', $line, 2);
				@list($key, $value) = preg_split('/:( |$)/', $line, 2);
				$head[] = array(ucwords(strtolower($key), '-'), @iconv_mime_decode($value, 0, 'UTF-8'));
			}
		}
		foreach ($head as list($key, $value))
			$head[$key] = $value;
		
		$origbody = $body = '';
		if ($withbody) {
			for (++$i; $i < count($lines); ++$i)
				$body .= $lines[$i] . "\r\n";
			$body = mymimebody($head, $origbody = $body, $dataurl);
		}
		return array($head, $body, $origbody);
	}
	
	$liste = array();
	$nomid = 0;
	
	$single = null;
	
	$txtfp = fopen($mbox, 'r');
	if (flock($txtfp, LOCK_SH)) {
		
		$fp = fopen($idx, 'r+');
		@list(, $size) = unpack('J', fread($fp, 8));
		if (!$size) {
			$size = 0;
			ftruncate($fp, 8);
		}
		fseek($fp, 8);
		
		fseek($txtfp, 0, SEEK_END);
		if ($size < ftell($txtfp)) {
			fseek($txtfp, $size);
			
			$froms = array();
			$from = '';
			while (($line = fgets($txtfp)) !== false) {
				if (substr($line, 0, 5) === 'From ') {
					unset($from);
					$from = '';
					$froms[] =& $from;
				}
				$from .= $line;
			}
			unset($from);
			
			$pos = $size;
			
			fseek($fp, 0, SEEK_END);
			foreach ($froms as $i => $from) {
				$len = strlen($from);
				list($head) = myparsemime(myfrom2mime($from));
				fputcsv($fp, array(@$head['Message-Id'], @$head['From'], @$head['Subject'], @$head['Date'], $pos, $len));
				$pos += $len;
			}
			fseek($fp, 0);
			fwrite($fp, pack('J', ftell($txtfp)));
		}
		
		fseek($fp, 8);
		while (($line = fgetcsv($fp)) !== false) {
			@list($mid, , , , , ) = $line;
			if (!$line[2])
				$line[2] = '[No Subject]';
			$liste[$mid ?: $nomid++] = $line;
		}
		
		fclose($fp);
		
		if ($id = @$_GET['id']) {
			@list($mid, , , , $pos, $len) = $liste[$id];
			fseek($txtfp, (int)$pos);
			$from = fread($txtfp, (int)$len);
			$single = myparsemime(myfrom2mime($from), true);
			$single[0]['x-phil-size'] = philsize($len);
		}
	}
	
	flock($txtfp, LOCK_UN);
	fclose($txtfp);
	
	function mailtourl($head, $body) {
		$args = array('to' => @$head['Reply-To'] ?: @$head['From'],
					  'subject' => 'Re: ' . @$head['Subject'],
					  'body' => "\n\n" . preg_replace('@^@m', '> ', trim(htmlspecialchars_decode(strip_tags($body)))),
					  'in-reply-to' => @$head['Message-Id']);
		//return 'mailto:?' . http_build_query($args, encoding_type: PHP_QUERY_RFC3986);
		return 'send.php?' . http_build_query($args, encoding_type: PHP_QUERY_RFC3986);
	}
?>
<!doctype html>
<meta charset=utf-8 />
<meta content=width=device-width name=viewport />
<style>
	div.article img {
		width: 100%;
	}
	td[nowrap] {
		white-space: nowrap;
	}
</style>
<?php if (!isset($_GET['frame'])): ?>
<link href=style/sheet.css rel=stylesheet />
<title>Courriel</title>
<form method=post>
	<p><button name=_logout style="float: right; ">Logout</button></p>
</form>
<h1>Courriel</h1>
<?php endif; ?>
<?php if (list($head, $body, $origbody) = $single): ?>
<table border=1 cellpadding=10 width=100%>
<?php	foreach (array(array('From', 'From'),
					   array('To', 'To'),
					   array('Date', 'Date'),
					   array('Subject', 'Subject'),
					   array('Size', 'x-phil-size')) as list($name, $key)): ?>
	<tr>
		<td align=right nowrap width=1><?= htmlspecialchars($name) ?></td>
		<td align=left><?= htmlspecialchars(@$head[$key]) ?></td>
	</tr>
<?php	endforeach; ?>
</table>
<?= $body ?>
<?php	if ($has_html && !isset($_GET['html'])): ?>
<p><a href="<?= htmlspecialchars('?id=' . rawurlencode($id) . '&html' . $suffixsent) ?>">View HTML</a>
<hr />
<?php	endif; ?>
<?php	if ($file !== 'sent' && !isset($_GET['html'])): ?>
<p><a href="<?= htmlspecialchars(mailtourl($head, $body)) ?>">Reply</a>
<hr />
<?php	endif; ?>
<?php	if (isset($_GET['head'])): ?>
<table border=1 cellpadding=10 width=100%>
<?php		foreach (array_filter($head, 'is_array') as list($key, $value)): ?>
	<tr>
		<td align=right nowrap width=1><?= htmlspecialchars($key) ?></td>
		<td align=left style=word-break:break-all><?= htmlspecialchars($value) ?></td>
	</tr>
<?php		endforeach; ?>
</table>
<?php	elseif (isset($_GET['body'])): ?>
<pre class=article>
<?php
	$lines = array();
	foreach (array_filter($head, 'is_array') as list($key, $value)) {
		if (substr($key, 0, 8) === 'Content-')
			$lines[] = $key . ': ' . $value;
	}
	echo htmlspecialchars(implode("\n", $lines));
?>
</pre>
<pre class=article>
<?= htmlspecialchars($origbody) ?>
</pre>
<?php	else: ?>
<p>
	<a href="<?= htmlspecialchars('?id=' . rawurlencode($id) . '&head' . $suffixsent) ?>">Headers</a>
	<a href="<?= htmlspecialchars('?id=' . rawurlencode($id) . '&body' . $suffixsent) ?>">Body</a>
<?php	endif; ?>
<?php else: ?>
<?php	if (!isset($_GET['frame'])): ?>
<table width=100%><tr>
	<td align=left valign=bottom><a href=send.php>Send e-mail</a></td>
	<td align=right>
<?php		foreach ($files as $i => $f): $t = $f === 'zzzsent' ? 'sent' : basename($f); ?>
<?php			if ($i): ?>
		-
<?php			endif; ?>
<?php			if ($f === $file): ?>
		<a><?= htmlspecialchars($t) ?></a>
<?php			else: ?>
		<a href="<?= htmlspecialchars('?file=' . rawurlencode($f)) ?>"><?= htmlspecialchars($t) ?></a>
<?php			endif; ?>
<?php		endforeach; ?>
	</td>
</tr></table>
<?php	endif; ?>
<table border=1 cellpadding=10 width=100%>
<?php	if (isset($_GET['frame'])): ?>
	<tr>
		<th nowrap>Objet / De</th>
		<th nowrap>Date</th>
		<th nowrap>Taille</th>
	</tr>
<?php	endif; ?>
<?php	foreach (array_reverse($liste, true) as $mid => list(, $from, $subject, $date, , $len)): ?>
	<tr>
<?php		if (isset($_GET['frame'])): ?>
		<td>
			<a href="<?= htmlspecialchars('?id=' . rawurlencode($mid) . $suffixsent) ?>" target=single><?= htmlspecialchars($subject) ?></a>
			<br />
			<span><?= htmlspecialchars($from) ?></span>
		</td>
		<td nowrap width=1><?= htmlspecialchars($date) ?></td>
		<td nowrap width=1><?= htmlspecialchars(philsize($len)) ?></td>
<?php		else: ?>
		<td align=left>
			<span><?= htmlspecialchars($from) ?></span>
			<br />
			<a href="<?= htmlspecialchars('?id=' . rawurlencode($mid) . $suffixsent) ?>"><?= htmlspecialchars($subject) ?></a>
			<br />
			<span><?= htmlspecialchars($date) ?></span>
			<br />
			<span><?= htmlspecialchars(philsize($len)) ?></span>
		</td>
<?php		endif; ?>
	</tr>
<?php	endforeach; ?>
</table>
<?php endif; ?>
